open F

type type_error =
  | TypeMismatch of debruijn_type * debruijn_type
  | ArityMismatch of arity_requirement * debruijn_type
  | NotAnArrow of debruijn_type
  | NotAProduct of debruijn_type
  | NotAForall of debruijn_type
  | UnboundTermVariable of tevar
  | TwoOccurencesOfSameVariable of string

and arity_requirement = Index of int | Total of int

exception TypeError of type_error

(* -------------------------------------------------------------------------- *)

(* A type environment maps term variables to types in de Bruijn's representation. *)

(* Our term variables are in nominal representation, which allows us to represent
   the environment as a binary search tree. One difficulty is that when we enter a
   new type variable binder [TyAbs], we must (conceptually) shift every type in
   the range of the environment up by one. When term variables are in de Bruijn
   notation and the environment is a list, this is easily done by inserting a mark
   in front of the environment. Here, we must be more clever. We maintain an [N2DB]
   environment that tells us, for every *term* variable, how long ago it was bound,
   that is, how many *type* variables were introduced since then. Hence, this tells
   us by how much we must shift its type. It's really a bit too clever, but let's
   do this for fun. *)

module TermVar =
  String

module TermVarMap =
  Map.Make(TermVar)

module N2DB =
  DeBruijn.Nominal2deBruijn(TermVar)

type env = {
  (* A mapping of term variables to types. *)
  types: debruijn_type TermVarMap.t;
  (* A translation environment of TERM variables to TYPE indices. *)
  names: N2DB.env;
}

type _datatype_env = (unit, debruijn_type) Datatype.Env.t

let empty =
  { types = TermVarMap.empty; names = N2DB.empty }

let lookup { types; names } x =
  try
    (* Obtain the type associated with [x]. *)
    let ty = TermVarMap.find x types in
    (* Find how long ago [x] was bound. *)
    let w = N2DB.lookup names x in
    (* Shift the type [ty] by this amount, so that it makes sense in the
       current scope. *)
    Type.lift w 0 ty
  with Not_found ->
    (* must have been raised by [TermVarMap.find] *)
    raise (TypeError (UnboundTermVariable x))

let extend_with_tevar { types; names } x ty =
  (* Map the name [x] to [ty], and record when it was bound, without
     incrementing the time. *)
  { types = TermVarMap.add x ty types;
    names = N2DB.slide names x }

let extend_with_tyvar { types; names } =
  (* Increment the time. *)
  { types; names = N2DB.bump names }

let singleton x ty =
  extend_with_tevar empty x ty

let concat env1 env2 =
  let combine _ _ ty2 = Some ty2 in
  { types = TermVarMap.union combine env1.types env2.types ;
    names = N2DB.concat env1.names env2.names }

let concat_disjoint env1 env2 =
  let exception Conflict of TermVarMap.key in
  try
    let combine x _ _ =
      raise (Conflict x) in
    let types = TermVarMap.union combine env1.types env2.types
    and names = N2DB.concat env1.names env2.names in
    Ok { types; names; }
  with Conflict x -> Error x

(* -------------------------------------------------------------------------- *)

(* Destructors. *)

let unfold ty =
  match ty with
  | TyMu ((), body) ->
      Type.subst ty 0 body
  | _ ->
      assert false

let rec as_arrow ty =
  match ty with
  | TyArrow (ty1, ty2) ->
      ty1, ty2
  | TyMu _ ->
      as_arrow (unfold ty)
  | _ ->
      raise (TypeError (NotAnArrow ty))

let rec as_product ty =
  match ty with
  | TyProduct tys ->
      tys
  | TyMu _ ->
      as_product (unfold ty)
  | _ ->
      raise (TypeError (NotAProduct ty))

let rec as_forall ty =
  match ty with
  | TyForall ((), ty) ->
      ty
  | TyMu _ ->
      as_forall (unfold ty)
  | _ ->
      raise (TypeError (NotAForall ty))

(* -------------------------------------------------------------------------- *)

(* [equal1] groups the easy cases of the type equality test. It does not
   accept arguments of the form [TyMu]. It expects a self parameter [equal]
   that is used to compare the children of the two root nodes. *)

let pointwise equal tys1 tys2 =
  try
    List.for_all2 equal tys1 tys2
  with Invalid_argument _ ->
    (* Catching [Invalid_argument] is bad style, but saves a couple calls
       to [List.length]. A cleaner approach would be to use arrays instead
       of lists. *)
    false

let equal1 equal ty1 ty2 =
  match ty1, ty2 with
  | TyMu _, _
  | _, TyMu _ ->
      assert false
  | TyVar x1, TyVar x2 ->
      Int.equal x1 x2
  | TyArrow (ty1a, ty1b), TyArrow (ty2a, ty2b) ->
      equal ty1a ty2a && equal ty1b ty2b
  | TyForall ((), ty1), TyForall ((), ty2) ->
      equal ty1 ty2
  | TyProduct tys1, TyProduct tys2 ->
      pointwise equal tys1 tys2
  | TyConstr (Datatype.Type tyconstr1, tys1),
    TyConstr (Datatype.Type tyconstr2, tys2) ->
      String.equal tyconstr1 tyconstr2 &&
      pointwise equal tys1 tys2
  | (TyVar _ | TyArrow _ | TyForall _ | TyProduct _ | TyConstr _),
    _ ->
      false

(* -------------------------------------------------------------------------- *)

(* An unsound approximation of the equality test. (Now unused.) *)

(* Comparing two [TyMu] types simply by comparing their bodies would be
   correct, but incomplete: some types that are equal would be considered
   different. *)

(* Naively unfolding every [TyMu] type is sound and complete, but leads to
   potential non-termination. The following code avoids non-termination by
   imposing a budget limit. This makes this equality test unsound: two types
   may be considered equal when they are in fact different. *)

let rec equal budget ty1 ty2 =
  match ty1, ty2 with
  | (TyMu _ as ty1), ty2
  | ty2, (TyMu _ as ty1) ->
      budget = 0 || equal (budget - 1) (unfold ty1) ty2
  | _, _ ->
      equal1 (equal budget) ty1 ty2

let budget =
  4

let _unsound_equal ty1 ty2 =
  equal budget ty1 ty2

(* -------------------------------------------------------------------------- *)

(* A sound approximation of the equality test. *)

(* The System F types produced by an ML type inference algorithm cannot
   contain [TyForall] under [TyMu]. In this restricted setting, deciding
   type equality is relatively easy. We proceed in two layers, as follows:

   - Initially, we perform a normal (structural) equality test, going
     down under [TyForall] constructors if necessary. As soon as we
     hit a [TyMu] constructor on either side, we switch to the lower
     layer:

   - In the lower layer, we know that we cannot hit a [TyForall]
     constructor (if we do, we report that the types differ). We
     map each type (separately) to a graph of unification variables
     (where every variable has rigid structure), then we check that
     the two graphs can be unified. *)

exception Clash =
  Structure.Iter2

module S =
  Structure

type vertex =
  structure UnionFind.elem

and structure =
  | SRigidVar of DeBruijn.index
  | SStruct of vertex S.structure

let rec translate (env : vertex list) (ty : F.debruijn_type) : vertex =
  match ty with

  | TyVar x ->
      (* If [x] is bound in [env], then [x] is bound by μ, and we
         already have a vertex that represents it. Otherwise, [x]
         is bound by ∀, and it is regarded as a rigid variable. *)
      let n = List.length env in
      if x < n then
        List.nth env x
      else
        UnionFind.make (SRigidVar (x - n))

  | TyMu ((), ty) ->
      (* Extend the environment with a mapping of this μ-bound variable
         to a fresh vertex. *)
      let dummy : structure = SRigidVar 0 in
      let v1 = UnionFind.make dummy in
      let env = v1 :: env in
      (* Translate the body. *)
      let v2 = translate env ty in
      (* Unify the vertices [v1] and [v2], keeping [v2]'s structure. *)
      let v = UnionFind.merge (fun _ s2 -> s2) v1 v2 in
      v

  | TyForall _ ->
      (* Either we have found a ∀ under a μ, which is unexpected, or one
         of the types that we are comparing has more ∀ quantifiers than
         the other. Either way, we report that the types differ. *)
      raise Clash

  | TyArrow (ty1, ty2) ->
      let v1 = translate env ty1
      and v2 = translate env ty2 in
      UnionFind.make (SStruct (S.TyArrow (v1, v2)))

  | TyProduct tys ->
      let vs = translate_list env tys in
      UnionFind.make (SStruct (S.TyProduct vs))

  | TyConstr (id, tys) ->
      let vs = translate_list env tys in
      UnionFind.make (SStruct (S.TyConstr (id, vs)))

and translate_list env tys =
  List.map (translate env) tys

let insert q v1 v2 =
  Stack.push (v1, v2) q

let unify_struct q s1 s2 =
  match s1, s2 with
  | SRigidVar x1, SRigidVar x2 ->
      if x1 <> x2 then raise Clash
  | SStruct s1, SStruct s2 ->
      Structure.iter2 (insert q) s1 s2
  | SRigidVar _, SStruct _
  | SStruct _, SRigidVar _ ->
      raise Clash

let merge_struct q s1 s2 =
  unify_struct q s1 s2;
  s1

let rec unify q (v1 : vertex) (v2 : vertex) =
  let (_ : vertex) = UnionFind.merge (merge_struct q) v1 v2 in
  unify_pending q

and unify_pending q =
  match Stack.pop q with
  | v1, v2 ->
      unify q v1 v2
  | exception Stack.Empty ->
      ()

let unify v1 v2 =
  let q = Stack.create() in
  unify q v1 v2

let rec equal ty1 ty2 =
  match ty1, ty2 with
  | (TyMu _ as ty1), ty2
  | ty2, (TyMu _ as ty1) ->
      (* We have hit a μ binder on one side. Switch to layer-1 mode. *)
      begin try
        let env = [] in
        let v1 = translate env ty1
        and v2 = translate env ty2 in
        unify v1 v2;
        true
      with Clash ->
        false
      end
  | _, _ ->
      (* Otherwise, continue in layer-2 mode. *)
      equal1 equal ty1 ty2

(* -------------------------------------------------------------------------- *)

(* Re-package the type equality test as a type equality check. *)

let (--) ty1 ty2 : unit =
  if not (equal ty1 ty2) then
    raise (TypeError (TypeMismatch (ty1, ty2)))

(* -------------------------------------------------------------------------- *)

(* The type-checker. *)

let nth_type ty tys i =
  if i < 0 || i >= List.length tys then
    raise (TypeError (ArityMismatch (Index i, ty)))
  else
    List.nth tys i

let rec typeof datatype_env env (t : debruijn_term) : debruijn_type =
  let typeof = typeof datatype_env in
  match t with
  | Var (_, x) ->
      lookup env x
  | Hole (_, ty, ts) ->
      let on_subterm t =
        ignore (typeof env t) in
      List.iter on_subterm ts;
      ty
  | Annot (_, t, ty) ->
      typeof env t -- ty;
      ty
  | Abs (_, x, ty1, t) ->
      let ty2 = typeof (extend_with_tevar env x ty1) t in
      TyArrow (ty1, ty2)
  | App (_, t, u) ->
      let ty1, ty2 = as_arrow (typeof env t) in
      typeof env u -- ty1;
      ty2
  | Let (_, x, t, u) ->
      let env = extend_with_tevar env x (typeof env t) in
      typeof env u
  | TyAbs (_, (), t) ->
      TyForall ((), typeof (extend_with_tyvar env) t)
  | TyApp (_, t, ty2) ->
      Type.subst ty2 0 (as_forall (typeof env t))
  | Tuple (_, ts) ->
      let tys = List.map (typeof env) ts in
      TyProduct tys
  | Proj (_, i, t) ->
      let ty = typeof env t in
      let tys = as_product ty in
      nth_type ty tys i
  | Variant (_, lbl, dty, t) ->
      let arg_type = typeof_variant datatype_env lbl dty in
      let ty = typeof env t in
      ty -- arg_type;
      TyConstr dty
  | LetProd (_, xs, t, u) ->
      let ty = typeof env t in
      let tys = as_product ty in
      if List.length xs <> List.length tys then
        raise (TypeError (ArityMismatch (Total (List.length xs), ty)));
      let env = List.fold_left2 extend_with_tevar env xs tys in
      typeof env u
  | Match (_, ty, t, brs) ->
      let scrutinee_ty = typeof env t in
      let tys = List.map (typeof_branch datatype_env env scrutinee_ty) brs in
      List.iter ((--) ty) tys;
      ty

and typeof_variant datatype_env lbl (tid, tyargs) =
  let Datatype.{ data_kind ; type_params ; labels_descr; _ } =
    Datatype.Env.find_decl datatype_env tid in
  let Datatype.{ arg_type; _ } =
    List.find (fun l -> l.Datatype.label_name = lbl) labels_descr in
  begin
    match data_kind with
    | Datatype.Variant ->
        ()
    | Datatype.Record ->
        raise Datatype.Env.UnexpectedRecord
  end;
  List.fold_right2 (fun ty () arg_type ->
    Type.subst ty 0 arg_type
  ) tyargs type_params arg_type

and typeof_branch datatype_env env scrutinee_ty (pat, rhs) =
  let new_bindings = binding_pattern datatype_env scrutinee_ty pat in
  let new_env = concat env new_bindings in
  typeof datatype_env new_env rhs

and binding_pattern datatype_env scrutinee_ty pat =
  let binding_pattern = binding_pattern datatype_env in
  match pat with
  | PVar (_, x) ->
      singleton x scrutinee_ty
  | PWildcard _ ->
      empty
  | PAnnot (_, pat, ty) ->
      scrutinee_ty -- ty;
      binding_pattern ty pat
  | PTuple (_, ps) ->
      let tys = as_product scrutinee_ty in
      let envs = List.map2 binding_pattern tys ps in
      let f env1 env2 =
        match concat_disjoint env1 env2 with
        | Ok env -> env
        | Error x -> raise (TypeError (TwoOccurencesOfSameVariable x))
      in
      List.fold_left f empty envs
  | PVariant (_, lbl, dty, pat) ->
      let arg_type = typeof_variant datatype_env lbl dty in
      scrutinee_ty -- TyConstr dty;
      binding_pattern arg_type pat

let typeof datatype_env t =
  typeof datatype_env empty t

let typeof_result datatype_env u =
  match typeof datatype_env u with
  | v -> Ok v
  | exception TypeError e ->
    Error e
