(******************************************************************************)
(*                                                                            *)
(*                                  Inferno                                   *)
(*                                                                            *)
(*                       François Pottier, Inria Paris                        *)
(*                                                                            *)
(*  Copyright Inria. All rights reserved. This file is distributed under the  *)
(*  terms of the MIT License, as described in the file LICENSE.               *)
(*                                                                            *)
(******************************************************************************)

open Signatures

(**This module offers a set of combinators for building {b constraints} as
   well as a {b constraint solver}. This allows the user to express a type
   inference problem as a constraint. Then, {b type inference} amounts to
   solving this constraint. Naturally, the constraint solver produces a
   Boolean outcome: whether the constraint is satisfiable or unsatisfiable.
   Furthermore, in case the constraint is satisfiable, it also produces a {b
   semantic value}, an OCaml value that is computed in a bottom-up fashion.
   This allows the user to perform {b elaboration}, that is, to produce an
   explicitly-typed program as an outcome.

   The parameter [X] provides the {i scheme variables} that are bound in [def]
   and [let] constraints and are referred to in instantiation constraints.

   The parameter [S] provides the {i shallow structure} that the unification
   algorithm must use.

   The parameter [O] provides the {i decoded types} that are used during
   elaboration in order to construct an explicitly-typed program. *)
module Make
  (X : sig (** @inline *) include TEVAR end)
  (S : sig (** @inline *) include HSTRUCTURE end)
  (O : sig (** @inline *) include OUTPUT
                                  with type 'a structure = 'a S.structure end)
: sig

  (**A scheme variable [x] is bound by {!def}, {!let1}, {!letn}
     and is referred to by {!instance}. Such a variable denotes
     a Hindley-Milner type scheme σ, or (equivalently) a constraint
     abstraction λα.c. *)
  type tevar = X.t

  (* The type ['a structure] of shallow types is provided by [S]
     (and repeated by [O]). *)

  (* The types [tyvar] and [ty] of decoded type variables and decoded types
     are provided by the module [O]. *)
  open O

  (**A type variable α denotes an unknown type. The existential quantifier
     {!exist} introduces a {i flexible} type variable, whose value can be
     chosen so as to satisfy a constraint. The combinators {!(--)} and
     {!(---)} allow imposing equality constraints on type variables. *)
  type variable

  (**A shallow type is a piece of structure whose children are
     type variables. *)
  type shallow_ty =
    variable structure

  (**A deep type is a either a type variable or a piece of structure
     whose children are again deep types. *)
  type deep_ty =
    | DeepVar of variable
    | DeepStructure of deep_ty structure

  (**A decoded scheme consists of a list of quantifiers and a body.
     The quantifiers are decoded type variables; the body is a decoded type.
     The combinators {!let1} and {!letn} produce schemes as part of
     their semantic value. *)
  type scheme = tyvar list * ty

  (* ---------------------------------------------------------------------- *)

  (** {2 Building Constraints} *)

  (**['a co] is the type of a constraint whose resolution produces a semantic
     value of type ['a]. One can think of a constraint as a program expressed
     in a limited domain-specific language. Then, a constraint of type ['a co]
     is a program whose execution, if successful, produces a result of type
     ['a]. *)
  type 'a co

  (* ---------------------------------------------------------------------- *)

  (* The applicative functor interface. *)

  (* The following three operations form the applicative functor interface.
     [pure] represents the true constraint, while [(and+)] builds a conjunction
     of constraints. [(let+)] does not per se produce any constraint; it only
     performs postprocessing, i.e., it applies a user-supplied function to
     the value that is obtained after constraint solving. *)

  (**[pure x] is a constraint whose resolution always succeeds and produces
     the semantic value [x]. In other words, it is a {i true} constraint. *)
  val pure: 'a -> 'a co

  (**The binding form [let+] names the semantic value produced by a
     constraint. This allows embedding a semantic action (a computation
     expressed as a piece of OCaml code) in a constraint. This code plays no
     role in determining whether the constraint is satisfiable. It is
     executed, after the constraint has been found to be satisfiable, as part
     of the bottom-up elaboration phase. It transforms a semantic value of
     type ['a] into a semantic value of type ['b]. For example, a [map]
     combinator can be defined as follows:
     {[
       let map f c =
         let+ x = c in
         f x
     ]} *)
  val (let+): 'a co -> ('a -> 'b) -> 'b co

  (**The binding form [and+] can be used, together with [let+], to name the
     semantic values produced by several constraints. This has the effect of a
     logical conjunction: the composite constraint is satisfied when each of
     the component constraints is satisfied. For example, a binary
     [conjunction] combinator can be defined as follows:
     {[
       let conjunction c1 c2 =
         let+ x1 = c1
         and+ x2 = c2 in
         (x1, x2)
     ]} *)
  val (and+): 'a co -> 'b co -> ('a * 'b) co

  (* ---------------------------------------------------------------------- *)

  (* Equations. *)

  (**The constraint [α -- β] imposes an equality between the type variables
     [α] and [β]. Its semantic value is a unit value [()]. *)
  val (--): variable -> variable -> unit co

  (**The constraint [α --- τ] imposes an equality between the type variable
     [α] and the shallow type [τ]. The combinator [(---)] can be defined in
     terms of the combinators {!lift} and {!(--)}: it is just [lift (--)]. *)
  val (---): variable -> shallow_ty -> unit co

  (* ---------------------------------------------------------------------- *)

  (* Introducing type variables. *)

  (**A binder of type [('var, 'a) binder] introduces an object of type ['var]
     in the construction of a constraint of type ['a co]. For example, the
     binder {!exist} introduces a fresh type variable [α] and binds it
     existentially.

     A binder can be used in two equivalent styles. in indirect style, one
     writes [exist (fun α -> c)], where [c] is a constraint. In direct style,
     one writes [let@ α = exist in c]. In either case, the result is the
     constraint [∃α.c]. *)
  type ('var, 'a) binder =
    ('var -> 'a co) -> 'a co

  (**The binding form [let@ α = binder in c] gives the name α to the object
     introduced by the binder [binder] in the construction of the constraint
     [c]. *)
  val (let@) : ('a, 'r) binder -> ('a -> 'r co) -> 'r co

  (**The binder [exist] creates a fresh type variable and binds it
     existentially. *)
  val exist: (variable, 'r) binder

  (**The binder [shallow τ] is analogous to [exist], but additionally
     constrains the new type variable to be equal to the shallow type [τ].
     Thus, it offers a way of constructing a variable that stands for
     a shallow type.
     [let@ α = shallow τ in c] is equivalent to
     {[
       let@ α = exist in
       let+ () = α --- τ
       and+ v = c in
       v
     ]} *)
  val shallow: shallow_ty -> (variable, 'r) binder

  (**If [f] is a binary predicate whose second argument is a type variable,
     then [lift f] is a binary predicate whose second argument is a shallow
     type. [lift f α τ] is equivalent to [let@ β = shallow τ in f α β]. *)
  val lift: ('a -> variable -> 'b co) -> 'a -> shallow_ty -> 'b co

  (**The binder [deep θ] is analogous to [exist], but additionally constrains
     the new type variable to be equal to the deep type [θ]. Thus, it offers a
     way of constructing a variable that stands for a deep type. *)
  val deep: deep_ty -> (variable, 'r) binder

  (* ---------------------------------------------------------------------- *)

  (**The constraint [decode α] is always satisfied. Its semantic value is the
     decoded type that corresponds to the type variable [α]. This decoded type
     is computed after constraint solving has succeeded. If the type variable
     [α] is existentially bound, then the decoded type produced by [decode α]
     is the witness, that is, the value that must be assigned to [α] for the
     constraint to be satisfied. *)
  val decode: variable -> ty co

  (* ---------------------------------------------------------------------- *)

  (* Application of constraint abstractions, also known as instantiation. *)

  (**The constraint [instance x α] requires the type variable [α] to be an
     instance of the scheme denoted by the variable [x]. The variable [x]
     must have been bound earlier by {!def}, {!let1}, or {!letn}.

     On paper, this constraint is often written [x ≤ α]. One can also think of
     [x] as a constraint abstraction, in which case this constraint can be
     written [x(α)], still on paper.

     The semantic value of the constraint [instance x α] is a list of decoded
     types. This list indicates how the quantifiers of the scheme denoted by
     [x] have been instantiated. The length and order of this list match the
     length and order of the quantifiers of the scheme denoted by [x]. This
     scheme is obtained as part of the semantic value of the combinator
     {!let1} and {!letn} where [x] is bound. *)
  val instance: tevar -> variable -> ty list co

  (* ---------------------------------------------------------------------- *)

  (* Construction of constraint abstractions, also known as generalization. *)

  (**The constraint [def x α c] binds the variable [x] to the type variable
     [α] (viewed as a trivial monomorphic scheme) in the constraint [c].
     Under this binding, an instantiation constraint [instance x β] is
     equivalent to the equality constraint [α -- β]. *)
  val def: tevar -> variable -> 'a co -> 'a co

  (* [let1 x c1 c2] binds the term variable [x] to the constraint abstraction
     [c1] in the constraint [c2]. (Technically, [c1] is a function of a fresh
     type variable to a constraint, as in [exist].) The resulting constraint
     produces a tuple of 4 results, namely:
     - a list [vs] of (decoded) type variables that may appear in [a1], and (hence)
       should be universally bound by the client in [a1]. In a typical scenario,
       [a1] is a System F term, and the client will build the type abstraction
       [\Lambda vs.a1].
     - the (decoded) type scheme that was assigned to [x] while solving [c2].
     - the value [a1] produced by the constraint [c1].
     - the value [a2] produced by the constraint [c2]. *)
  val let1: tevar -> (variable -> 'a co) -> 'b co ->
            (tyvar list * scheme * 'a * 'b) co

  (* [let0 c] has the same meaning as [c], but, like [let1], produces a list [vs]
     of the type variables that may appear in the result of [c]. *)
  val let0: 'a co  -> (tyvar list * 'a) co

  (* [letn xs c1 c2] binds [n] term variables [xs] to [n] constraint abstractions
     in the constraint [c2]. Here, [c1] is a function of [n] fresh type variables
     [vs] to a constraint. The [i]-th term variable, [x_i], ends up bound to the
     constraint abstraction of the [i]-th type variable in [c_1], which one could
     write [\lambda v_i.c_1]. *)
  val letn: tevar list -> (variable list -> 'a co) -> 'b co ->
            (tyvar list * scheme list * 'a * 'b) co

  (* [letr1 alphas x (fun alphavs v -> c1) c2] binds the scheme variable [x]
     to the constraint abstraction [fun v -> c1] in the contstraint [c2],
     where [c1] is also abstracted over a list of rigid variables, one for
     each client type variable passed in the list [alphas].

     The resulting constraint produces a tuple of 3 results, namely:
     - a list [vs] of (decoded) type variables corresponding to the client type
       variables [alphas].
     - the value [a1] produced by the constraint [c1].
     - the value [a2] produced by the constraint [c2].

     Note: unlike [let1], we do not necessarily decode the type scheme
     corresponding to [x]; it is less interesting as we know that it
     must use exactly the rigid variables [vs] and no other. The
     client could still compute the decoded scheme by asking to decode
     the witness type of [v] within [c1].
   *)
  val letr1: 'client_tyvar list
             -> tevar
             -> (('client_tyvar * variable) list -> variable -> 'a co)
             -> 'b co
             -> (tyvar list * 'a * 'b) co

  val letrn: 'client_tyvar list
             ->  tevar list
             -> (('client_tyvar * variable) list
                 -> (tevar * variable) list
                 -> 'a co)
             -> 'b co
             -> (tyvar list * (tevar * scheme) list * 'a * 'b) co
  (* ---------------------------------------------------------------------- *)

  (* Correlation with the source code. *)

  (**The type [range] describes a range in some source code. It is not
     interpreted by the solver. It is used only as part of error reports. *)
  type range =
    Lexing.position * Lexing.position

  (**The constraint [correlate range c] is equivalent to [c], but records that
     this constraint is correlated with the range [range] in the source code.
     This information is used in error reports. The exceptions {!Unbound},
     {!Unify}, and {!Cycle} carry a range: it is the nearest range that
     encloses the abstract syntax tree node where the error is detected. *)
  val correlate: range -> 'a co -> 'a co

  (** {2 Solving Constraints} *)

  (**This exception is raised by {!solve} when an instantiation constraint
     [instance x α] refers to an unbound variable [x]. *)
  exception Unbound of range * tevar

  (**This exception is raised by {!solve} when an equality constraint cannot
     be satisfied. It carries two decoded types, representing the equality
     constraint that could not be satisfied. *)
  exception Unify of range * ty * ty

  (**This exception is raised by {!solve} when a cyclic equality constraint
     is detected, if [rectypes] is [false]. It carries a decoded type, which
     is a recursive type: that is, this type involves a μ binder. *)
  exception Cycle of range * ty

  exception VariableScopeEscape of range

  (**[solve ~rectypes c] determines whether the constraint [c] is satisfiable,
     and if so, computes and returns its semantic value. (The constraint [c]
     must be closed, so it is satisfiable if and only if it is valid.) If [c]
     is ill-formed or unsatisfiable, one of the exceptions {!Unbound},
     {!Unify}, and {!Cycle} is raised. The parameter [rectypes] determines
     whether solutions that involve recursive types are permitted; {!Cycle}
     can be raised only when [rectypes] is [false].

     The constraint [c] must be a {i toplevel constraint}. This means that
     every (binding or free) occurrence of a type variable must appear in the
     left-hand side of {!let0}, {!let1}, or {!letn}. To turn an arbitrary
     constraint [c] into a toplevel constraint, just write [let0 c]. *)
  val solve: rectypes:bool -> 'a co -> 'a

  (**/**)

  (**[pprint] turns a constraint into a PPrint document. This printer is
     provided for debugging purposes. *)
  val pprint : 'a co -> PPrint.document

end
